@@ -13,19 +13,17 @@
 # right then
 
 use strict;
-use warnings;
 use lib '/usr/share/apt-cacher/';
 
 use Cwd;
-
-## Just for testing!
-#use File::Basename;
-#use lib dirname(Cwd::abs_path $0);
-
 use Fcntl qw/:DEFAULT :flock F_SETFD/;
 use Getopt::Long qw(:config no_ignore_case bundling pass_through);
 use Digest::SHA1;
 use HTTP::Date;
+use FreezeThaw qw(freeze thaw);
+use IO::Uncompress::AnyUncompress qw($AnyUncompressError);
+use IO::Compress::Bzip2;
+use IO::Compress::Gzip;
 
 my $configfile = '/etc/apt-cacher/apt-cacher.conf';
 my $nice_mode=0;
@@ -82,11 +80,13 @@ define_global_lockfile($globlockfile);
 
 # check whether we're actually meant to clean the cache
 if ( $cfg->{clean_cache} ne 1 ) {
-    warn "Maintenance disallowed by configuration item clean_cache\n";
+    printmsg("Maintenance disallowed by configuration item clean_cache\n");
     exit 0;
 }
-# change uid and gid if root
-if ($cfg->{user} && !$> or $cfg->{group} && !$)) {
+
+# change uid and gid if root and another user/group configured
+if ($cfg->{user} && $cfg->{user} !~ 'root' && !$> 
+    or $cfg->{group} && $cfg->{group} !~ 'root' && !$) =~ /^0/){
     printmsg("Invoked as root, changing to $cfg->{user}:$cfg->{group} and re-execing.\n");
     setup_ownership($cfg);
     # Rexec to ensure /proc/self/fd ownerships correct which are needed for red
@@ -228,26 +228,25 @@ sub pdiff {
 	printmsg ("$name size not latest, proceeding with patch\n");
     }
 
-    # Need to decompress to patch
-    my $cat = ($name=~/bz2$/ ? "bzcat" : ($name=~/gz$/ ? "zcat" : "cat"));
-
     open(my $lck, $name)|| die "Unable to open file $name for locking: $!";
     flock($lck, LOCK_SH);
-    open(my $listpipe, "-|", $cat, $name)|| die "Unable to open input pipe for $name: $!";
+
+    my $raw = new IO::Uncompress::AnyUncompress $name
+      or die "Decompression failed: $AnyUncompressError\n";
+
     open (my $tfh, "+>", undef)|| die "Unable to open temp file: $!";
 	
     printmsg "Reading $basename...\n";
-    while (<$listpipe>){
+    while (<$raw>){
+	last if $AnyUncompressError;
 	print $tfh $_;
 	$sha1->add($_);
     }
-    close($listpipe);
+    close($raw);
     flock($lck, LOCK_UN);
     close($lck,);
-    my $rstat =($? >> 8);
-    #    printmsg "Read status $rstat\n";
-    if ($rstat) {
-	warn "$name Read failed, aborting patch\n";
+    if ($AnyUncompressError) {
+	warn "$name Read failed: $AnyUncompressError. Aborting patch\n";
 	return;
     }
 	
@@ -331,20 +330,19 @@ sub pdiff {
 	    }
 	}
 	printmsg "Reading $pdiff\n";
-	open(PDIFF, "-|", 'zcat', $pdiff)|| die "Unable to open file $pdiff: $!";
-	flock(PDIFF, LOCK_EX);
-	while (<PDIFF>) {
+	my $pdfh = new IO::Uncompress::AnyUncompress $pdiff
+	  or die "Decompression failed: $AnyUncompressError\n";
+	while (<$pdfh>) {
+	    last if $AnyUncompressError;
 	    print DIFFS $_;
 	    $sha1->add($_);
 	}
-	flock(PDIFF, LOCK_UN);
-	close(PDIFF);
-	$rstat =($? >> 8);
-	#	printmsg "Read status $rstat\n";
-	if ($rstat) {
-	    warn "Read $pdiff failed, aborting patch\n";
+	close($pdfh);
+	if ($AnyUncompressError) {
+	    warn "$name Read failed: $AnyUncompressError. Aborting patch\n";
 	    return;
 	}
+
 	my $pdiffdigest = $sha1->hexdigest;
 #	printmsg "$pdiff SHA1: $pdiffdigest\n";
 	if ($pdiffsha1 ne $pdiffdigest) {
@@ -364,7 +362,7 @@ sub pdiff {
     print $patchpipe $diffs;
     close($patchpipe);
     chdir $cwd or die "Unable to restore working directory: $!"; # Restore
-    $rstat =($? >> 8);
+    my $rstat =($? >> 8);
     if ($rstat) {
 	warn "Patching failed (exit code $rstat), aborting\n";
 	return;
@@ -389,12 +387,16 @@ sub pdiff {
 		printmsg "Warning: $destfile already exists\n";
 	    }
 	    seek($tfh,0,0); # rewind
-	    my ($zip,$encoding) = ($name=~/bz2$/ ? ("bzip2","x-bzip2") : ($name=~/gz$/ ? ("gzip -9nc","x-gzip") : "cat"));
-	    open(my $writepipe, "| $zip > $destfile") || die "Unable to open output pipe: $!\n";
+	    my ($z,$encoding) = ($name=~/bz2$/ ?
+				 ((new IO::Compress::Bzip2 $destfile), "x-bzip2") :
+				 ($name=~/gz$/ ?
+				  ((new IO::Compress::Gzip $destfile, -Level => 9), "x-gzip") :
+				  new IO::Handle $destfile));
+
 	    while(<$tfh>) {
-		print $writepipe $_;
+		$z->print($_);
 	    }
-	    close($writepipe);
+	    close($z);
 	    my @info = stat($destfile);
 	    my $datestring = HTTP::Date::time2str;
 	    open(my $hfh, ">", "$destfile.header") || die "Unable to open header file: $!\n";
@@ -427,7 +429,7 @@ EOF
 	    # Read checksums
 	    if ($cfg->{checksum}) {
 		printmsg ("Importing new checksums from patched $name\n");
-		import_sums($tfh);
+		import_sums($tfh,&db);
 	    }
 	}
     }
@@ -440,9 +442,10 @@ EOF
 }
 
 # Calls _db_compact to do the work and reports results
+# Arg: DB handle ref
 sub db_compact {
     printmsg "Compacting checksum database....\n";
-    my ($status, $results) = &_db_compact;
+    my ($status, $results) = &_db_compact($_[0]);
     if ($$status) {
 	printmsg "db_compact failed: $$status\n";
     }
@@ -458,14 +461,14 @@ if (@db_mode || $db_recover){
     my $ok_chars = '-a-zA-Z0-9+_.~'; # Acceptable characters for user input
     print "Checksum database mode\n";
 
+    require 'apt-cacher-lib-cs.pl';
+
     if (!$cfg->{checksum}) {
-	die "checksumming not enabled. Use --force to override\n" if !$force;
-	print "checksumming not enabled, but forced to continue\n";
+	die "$0: checksumming not enabled. Use --force to override\n" if !$force;
+	print "$0: checksumming not enabled, but forced to continue\n";
     }
     $verbose = 1; # Just for now
 
-    require 'apt-cacher-lib-cs.pl';
-
     if ($db_recover) {
 	printmsg "Running database recovery...";
 	&db_recover;
@@ -474,19 +477,19 @@ if (@db_mode || $db_recover){
 
     chdir "$cfg->{cache_dir}/packages" || die "Unable to enter cache dir: $!";
 
-    my $dbref = &db_init("$cfg->{cache_dir}/sums.db")|| die "Unable to init db: $!\n";
+    my $dbh = ${&db}|| die "Unable to init db: $!\n";
 
   SWITCH:
     for (@db_mode) {
 	/^import/ && do {
 	    for (<*es.bz2>, <*es.gz>, <*es>, <*Release>, <*diff_Index>) {
 		printmsg "Importing checksums from $_\n";
-		&import_sums($_) if !$sim_mode;
+		&import_sums($_,\$dbh) if !$sim_mode;
 	    }
 	    next SWITCH;
 	};
 	/^compact/ && do {
-	    &db_compact;
+	    &db_compact(\$dbh);
 	    next SWITCH;
 	};
 	/^(?:dump|search)/ && do {
@@ -496,31 +499,37 @@ if (@db_mode || $db_recover){
 		die "No search expression given\n" if !$re;
 		die "Invalid character '$1' in search\n" if $re =~ /([^$ok_chars])/o; # sanitize
 	    }
-	    while (my ($file,$data) = each %$dbref){
-		next if /^search/ && $file !~ /$re/;
-		print "$file\n";
-		my $href = hashify(\$data);
-		while (my ($k,$v) = each %$href) {
-		    $v='' if ! defined $v;
-		    print " $k: $v\n";
-		}
-	    }
+	    my $cursor = get_cursor($dbh);
+	    my ($filename,$data)=('','');
+	    while (&cursor_next($cursor, \$filename, \$data) == 0)
+	      {
+		  next if /^search/ && $filename !~ /$re/;
+		  print "$filename\n";
+		  my $href = hashify(\$data);
+		  while (my ($k,$v) = each %$href) {
+		      $v='' if ! defined $v;
+		      print " $k: $v\n";
+		  }
+	      }
 	    next SWITCH;
 	};
 	/^delete/ && do {
 	    my $re=shift;
 	    die "No give regex to match files to delete\n" if !$re;
 	    die "Invalid character '$1' in pattern\n" if $re =~ /([^$ok_chars])/o; # sanitize
-	    while (my $file = each %$dbref){
-		next if $file !~ /$re/;
-		printmsg "Deleting data for $file\n";
-		delete $dbref->{$file} if !$sim_mode;
+	    my $cursor = get_cursor($dbh,1);
+	    my ($filename,$data)=('','');
+	    while (&cursor_next($cursor, \$filename, \$data) == 0)
+	      {
+		next if $filename !~ /$re/;
+		printmsg "Deleting data for $filename\n";
+		$cursor->c_del and warn "c_del failed: $!" if !$sim_mode;
 	    }
 	    next SWITCH;
 	};
 	/^verify/ && do {
 	    printmsg "Waiting for exclusive lock...";
-	    if (&db_lock(LOCK_EX)){
+	    if (&db_flock(LOCK_EX)){
 		printmsg "Got it!\nVerifying database...";
 		printmsg &db_verify("$cfg->{cache_dir}/sums.db", &temp_env) ? "Failed! $!\n" : "Passed!\n";
 	    }
@@ -540,17 +549,19 @@ if (@db_mode || $db_recover){
 
 # Cache cleaning from here
 
+# Take a lock on the cache dir to ensure only one cleanup script can run at a
+# time as this can take a while on some systems.
+
+open  (my $cleanup_lock, '<', $cfg->{'cache_dir'})
+  or die "Can't open $cfg->{'cache_dir'} for locking!\nError: $!\n";
+flock $cleanup_lock, LOCK_EX|LOCK_NB
+  or die "Another apt-cacher-cleanup is already running. Exiting!\n";
+
 # check offline mode in config
 if (defined $cfg->{offline_mode} && $cfg->{offline_mode}) {
 	$offline = 1;
 }
 
-my $dbref;
-if ($cfg->{checksum}) {
-   require 'apt-cacher-lib-cs.pl';
-   $dbref = &db_init("$cfg->{cache_dir}/sums.db");
-}
-
 use DB_File;
 tie my %valid, 'DB_File';
 
@@ -567,9 +578,26 @@ if($> == 0 && !$cfg->{user} && !$force) {
     die "Running $0 as root\nand no effective user has been specified. Aborting.\nPlease set the effective user in $configfile\n";
 }
 
+# Ensure corresponding Packages/Sources is present for each diff_Index
+DIFFINDEX:
+for (<*diff_Index>) {
+    if (/^(.+)\.diff_Index/) {
+	printmsg "Checking for $1 for $_\n";
+	for ($1,"$1.gz", "$1.bz2") {
+	    if ( -f $_ ) {
+		printmsg "Found $_\n";
+		next DIFFINDEX;
+	    }
+	}
+	printmsg ("Not found. Downloading\n");
+	# Might as well use bzipped files
+	&get("$1.bz2");
+    }
+}
+
 # file state decisions, lock that area
 set_global_lock(": file state decision");
-my @ifiles=(<*Release>, <*_Index>, <*es.gz>, <*es.bz2>, <*es>);
+my @ifiles=(<*Release>, <*_Index> , <*es.gz>, <*es.bz2>, <*es>);
 release_global_lock();
 
 for my $file (@ifiles) {
@@ -594,26 +622,6 @@ for my $file (@ifiles) {
    }
 }
 
-# Ensure corresponding Packages/Sources is present for each diff_Index
-DIFFINDEX:
-for (@ifiles) {
-    if (/^(.+).diff_Index/) {
-	printmsg "Checking for $1 for $_\n";
-	for ($1,"$1.gz", "$1.bz2") {
-	    if ($valid{$_}) {
-		printmsg "Found $_\n";
-		next DIFFINDEX;
-	    }
-	}
-	printmsg ("Not found. Downloading\n");
-	# Might as well use bzipped files
-	&get("$1.bz2");
-	push @ifiles, "$1.bz2";
-	$valid{"$1.bz2"}=1;
-    }
-}
-
-# use the list of config files we already know
 for my $file (@ifiles) {
     printmsg "Reading: $file\n";
 
@@ -653,17 +661,26 @@ printmsg "Found ".scalar (keys %valid)." valid file entries\n";
 
 # Remove old checksum data
 if ($cfg->{checksum}) {
-    my $do_compact;
-    $dbref && do {
-	  printmsg "Removing expired entries from checksum database\n";
-	  while (my $key = each %$dbref){
-	      next if defined $valid{$key};
-	      printmsg "Deleting checksum data for $key\n";
-	      delete $dbref->{$key} if !$sim_mode;
-	      $do_compact = 1;
-	  }
-	  &db_compact if $do_compact || $pdiff_mode;
-      };
+   require 'apt-cacher-lib-cs.pl';
+
+   my $dbh = ${&db};
+
+   my $do_compact;
+   $dbh && do {
+       printmsg "Removing expired entries from checksum database\n";
+
+
+       my $cursor = get_cursor($dbh,1);
+       my ($filename,$data)=('','');
+       while (&cursor_next($cursor, \$filename, \$data) == 0)
+	 {
+	   next if defined $valid{$filename};
+	   printmsg "Deleting checksum data for $filename\n";
+	   $cursor->c_del and warn "c_del failed: $!" if !$sim_mode;
+	   $do_compact = 1;
+       }
+       &db_compact(\$dbh) if $do_compact || $pdiff_mode;
+   };
 }
 
 for(<*.deb>, <*.udeb>, <*.bz2>, <*.gz>, <*.dsc>) {
@@ -695,7 +712,7 @@ for(<*.deb>, <*.udeb>, <*.bz2>, <*.gz>, <*.dsc>) {
 chdir "$cfg->{cache_dir}/headers" && -w "." || die "Could not enter the cache dir";
 
 # headers for previously expired files
-for(<*.deb>, <*.bz2>, <*.gz>, <*.dsc>) {
+for(<*.deb>, <*.udeb>, <*.bz2>, <*.gz>, <*.dsc>) {
    if(! defined($valid{$_})) {
       unlink $_, "../private/$_.complete" unless $sim_mode;
       printmsg "Removing expired headers: $_ and company...\n";
@@ -704,7 +721,7 @@ for(<*.deb>, <*.bz2>, <*.gz>, <*.dsc>) {
 
 # also remove void .complete files, created by broken versions of apt-cacher in rare conditions
 chdir "$cfg->{cache_dir}/private" && -w "." || die "Could not enter the cache dir";
-for(<*.deb.complete>, <*.bz2.complete>, <*.gz.complete>, <*.dsc.complete>) {
+for(<*.deb.complete>, <*.udeb.complete>, <*.bz2.complete>, <*.gz.complete>, <*.dsc.complete>) {
    s/.complete$//;
    if(! (defined($valid{$_}) && -e "../packages/$_" && -e "../headers/$_") ) {
       printmsg "Removing: $_.complete\n";
