@@ -12,14 +12,18 @@
 use strict;
 use warnings;
 use Fcntl qw/:flock/;
+use FreezeThaw qw(freeze thaw);
+use IO::Uncompress::AnyUncompress qw($AnyUncompressError);
 our $cfg;
 
 sub read_config {
     # set the default config variables
     my %config = (
-		  cache_dir => '/var/log/cache/apt-cacher',
+		  cache_dir => '/var/cache/apt-cacher',
 		  logdir => '/var/log/apt-cacher',
 		  admin_email => 'root@localhost',
+		  user => $>,
+		  group => (sub { my $g = $); $g =~ s/\s.*$//; $g; })->(),
 		  generate_reports => 0,
 		  expire_hours => 0,
 		  http_proxy => '',
@@ -32,7 +36,45 @@ sub read_config {
 		  allowed_hosts => '*',
 		  limit => 0,
 		  daemon_port => 3142,
-		  fetch_timeout => 300 # five minutes from now
+		  fetch_timeout => 300, # five minutes from now
+		  curl_idle_timeout => 120,
+		  index_files_regexp => '(?:' . join('|',
+						     ('Index',
+						      'Packages\.gz',
+						      'Packages\.bz2',
+						      'Packages',
+						      'Release',
+						      'Release\.gpg',
+						      'Sources\.gz',
+						      'Sources\.bz2',
+						      'Sources',
+						      'Contents-.+\.gz',
+						      'pkglist.*\.bz2',
+						      'release',
+						      'release\..*',
+						      'srclist.*\.bz2',
+						      'Translation-.+\.bz2',
+						     )
+						    ) . ')$',
+		  package_files_regexp => '(?:' . join('|',
+						       ('\.deb',
+							'\.rpm',
+							'\.dsc',
+							'\.tar\.gz',
+							'\.tar\.bz2',
+							'\.diff\.gz',
+							'\.udeb',
+							'index\.db-.+\.gz',
+							'\.jigdo',
+							'\.template'
+						       )
+						      ) .')$',
+		  installer_files_regexp => '^(?:' . join('|',
+							 ('vmlinuz',
+							  'linux',
+							  'initrd\.gz'
+							 )
+							) . ')$'
 		 );
 
     (my $config_file) = @_;
@@ -72,7 +114,7 @@ sub check_install() {
     # Die if we have not been configured correctly
     die "$0: No cache_dir directory!\n" if (!-d $cfg->{cache_dir});
 
-    my $uid = $cfg->{user}=~/^\d+$/ ? $cfg->{user} : getpwnam($cfg->{group});
+    my $uid = $cfg->{user}=~/^\d+$/ ? $cfg->{user} : getpwnam($cfg->{user});
     my $gid = $cfg->{group}=~/^\d+$/ ? $cfg->{group} : getgrnam($cfg->{group});
 
     if (!defined ($uid || $gid)) {
@@ -163,34 +205,35 @@ sub make_mask ($$) {
 
 # Arg is ref to flattened hash. Returns hash ref
 sub hashify {
-    unless ($cfg->{debug}){
-	no warnings 'uninitialized'
+    return undef unless ${$_[0]};
+    if (${$_[0]} =~ /^FrT;/) {
+	# New format: FreezeThaw
+	return (thaw($ {$_[0]}))[0];
+    } else {
+	# Old format: join
+	return {split(/ /, ${$_[0]})};
     }
-    return {split(/ /, ${$_[0]})};
 }
 
 # Stores data flattened for use in tied hashes
 sub extract_sums {
    my ($name, $hashref) = @_;
-   my ($cat, $listpipe, $indexbase);
+   my $indexbase;
 
-   # If arg is filehandle, fileno is undef
-   if (fileno $name) {
-       seek($name,0,0);
-       $listpipe = $name;
+   seek($name,0,0) if fileno($name);
+   my $raw = new IO::Uncompress::AnyUncompress $name
+     or die "Decompression failed: $AnyUncompressError\n";
+
+   # If arg is fd get filename
+   $name = readlink('/proc/self/fd/'.fileno($name)) if fileno($name);
+
+   if($name=~/^.*\/(.*_)(?:Index|Release)(?: \(deleted\))?$/) {
+       $indexbase=$1;
    }
-   else {
-       $cat = ($name=~/bz2$/ ? "bzcat" : ($name=~/gz$/ ? "zcat" : "cat"));
 
-       if($name=~/^(.*_)(?:Index|Release)$/) {
-	   $indexbase=$1;
-	   $indexbase=~s!.*/!!g;
-       }
-
-       open($listpipe, "-|", $cat, $name);
-   }
    my ($file,$complete,$sha1,$sha256,$md5,$size);
-   while(<$listpipe>) {
+   while(<$raw>) {
+       last if $AnyUncompressError;
        if(/^\s(\w{40})\s+(\d+)\s(\S+)\n/ && $indexbase) { # diff_Index/Release SHA1
 	   $sha1=$1;
 	   $size =$2;
@@ -237,22 +280,22 @@ sub extract_sums {
 	   $complete=1;
        }
        if(defined($file) && $complete) {
-	   $hashref->{$file}= join ' ', (
-					 size => $size || '',
-					 md5 => $md5 || '',
-					 sha1 => $sha1 || '',
-					 sha256 => $sha256 || '',
-					 stamp => time()
-				      );
+	   $hashref->{$file}= freeze {
+	       size => $size,
+		 md5 => $md5,
+		   sha1 => $sha1,
+		     sha256 => $sha256
+		   };
 	   for ($file, $size, $md5, $sha1, $sha256, $complete) {
 	       undef $_
 	   }
        }
    };
-   return if (fileno $name); # Leave caller to clear up if we were passed a filehandle
-   close($listpipe);
-   my $ret = ! ($? >> 8);
-   return $ret;
+   if ($AnyUncompressError) {
+       warn "$name Read failed: $AnyUncompressError. Aborting read\n";
+       return;
+   }
+   return 1;
 }
 
 
@@ -285,6 +328,7 @@ sub set_global_lock {
 sub release_global_lock {
     debug_message("Exiting critical section") if defined (&debug_message);
     flock($exlock, LOCK_UN);
+    close $exlock;
 }
 
 sub setup_ownership {
@@ -336,27 +380,16 @@ sub barf {
 	die "--- $0: Fatal: $errs\n";
 }
 
-my $index_files_regexp =
-  '('
-  . join('|', ('Index',
-	       'Packages.gz',
-	       'Packages.bz2',
-	       'Release',
-	       'Release.gpg',
-	       'Sources.gz',
-	       'Sources.bz2',
-	       'Contents-.+\.gz',
-	       'pkglist.*\.bz2',
-	       'release$',
-	       'release\..*',
-	       'srclist.*\.bz2',
-	       'Translation-.+\.bz2'
-	      )
-	)
-  . ')$';
-
 sub is_index_file {
-    return ($_[0] =~ /$index_files_regexp/);
+    return ($_[0] =~ /$cfg->{index_files_regexp}/);
+}
+
+sub is_package_file {
+    return ($_[0] =~ /$cfg->{package_files_regexp}/);
+}
+
+sub is_installer_file {
+    return ($_[0] =~ /$cfg->{installer_files_regexp}/);
 }
 
 ######### HOOKS ###########
@@ -367,10 +400,14 @@ sub import_sums {
 }
 
 # purpose: ?create?, lock the DB file and establish DB connection
-sub db_init {
+sub db {
    return 1;
 }
 
+# release tie
+sub db_disconnect {
+}
+
 # purpose: create hasher object
 sub data_init {
    return 1;
